!===============================================================================
!===============================================================================
! ELCIRC LINEAR SOLVER ROUTINES
!
! subroutine solve_jcg
! subroutine tridag
!
!===============================================================================
!===============================================================================

subroutine solve_jcg(itime,moitn,mxitn,rtol,s,x,b,bc,lbc)
!-------------------------------------------------------------------------------
! Jacobi Preconditioned Conjugate Gradient
!-------------------------------------------------------------------------------
#ifdef USE_MPIMODULE
  use mpi
#endif
  use elfe_glbl, only : rkind,np,npa,wtimer,iplg,ipgl,mnei,nnp,inp,errmsg
  use elfe_msgp
  implicit none
#ifndef USE_MPIMODULE
  include 'mpif.h'
#endif
  integer,intent(in) :: itime !for outputting only
  integer,intent(in) :: moitn,mxitn    !output interval and max iterations
  real(rkind),intent(in) :: rtol       !relative tolerance
  real(rkind),intent(in) :: s(np,0:mnei)  !sparse matrix
  real(rkind),intent(inout) :: x(npa)  !eta2 -- with initial guess
  real(rkind),intent(in) :: b(np)      !qel
  real(rkind),intent(in) :: bc(npa)    !b.c. (elbc)
  logical,intent(in) :: lbc(npa)       !b.c. flag
  integer :: itn,ip,jp,j
  real(rkind) :: rdotrl,rdotr,rdotzl,rdotz,old_rdotz,beta,alphal,alpha,cwtmp
  real(rkind) :: rtol2,rdotr0
  real(rkind) :: z(npa),r(npa),p(npa),sp(npa)
  integer :: inz(mnei,np),nnz(np)
  real(rkind) :: snz(0:mnei,np),bb(np)
!-------------------------------------------------------------------------------

! Load local storage for non-zeros of sparse matrix and include BCs
! Since the calcualtion is done for resident elements only, entire row of
! the global matrix will be contained in the local row of the matrix.
  nnz=0 !# of non-zero entries for each element
  do ip=1,np
    if(lbc(ip)) then !nnz(ip)=0: no non-zero entries
      if(bc(ip)<-9998) call parallel_abort('JCG: wrong b.c.')
      snz(0,ip)=1d0 !modified sparsem
      x(ip)=bc(ip)
      bb(ip)=bc(ip) !modified rrhs (qel)
    else
      snz(0,ip)=s(ip,0)
      bb(ip)=b(ip)
      do j=1,nnp(ip)
        jp=inp(ip,j)
        if(lbc(jp)) then
          if(bc(jp)<-9998) call parallel_abort('JCG: wrong b.c. (2)')
          bb(ip)=bb(ip)-s(ip,j)*bc(jp)
        else
          nnz(ip)=nnz(ip)+1
          inz(nnz(ip),ip)=jp !local node #
          snz(nnz(ip),ip)=s(ip,j)
        endif
      enddo
    endif
  enddo !i=1,np

! Initialization -- assume x is initial guess
#ifdef INCLUDE_TIMING
  cwtmp=mpi_wtime()
#endif
  call exchange_p2d(x) !update x ghost values
#ifdef INCLUDE_TIMING
  wtimer(7,2)=wtimer(7,2)+mpi_wtime()-cwtmp
#endif

  itn=0
  rdotrl=0d0
  do ip=1,np
    sp(ip)=snz(0,ip)*x(ip)
    do j=1,nnz(ip) 
      sp(ip)=sp(ip)+snz(j,ip)*x(inz(j,ip))
    enddo
    r(ip)=bb(ip)-sp(ip) !residual
    if(associated(ipgl(iplg(ip))%next)) then !interface node
      if(ipgl(iplg(ip))%next%rank<myrank) cycle !already in the sum so skip
    endif
    rdotrl=rdotrl+r(ip)*r(ip)
  enddo !ip

#ifdef INCLUDE_TIMING
  cwtmp=mpi_wtime()
#endif
  call exchange_p2d(r) !update r ghost values
  call mpi_allreduce(rdotrl,rdotr,1,rtype,MPI_SUM,comm,ierr)
#ifdef INCLUDE_TIMING
  wtimer(7,2)=wtimer(7,2)+mpi_wtime()-cwtmp
#endif

! Convergence based on square of 2norm
  rtol2=rtol*rtol
  rdotr0=rdotr !initial 2norm of residual
  if(myrank==0) then
    write(33,'(//a,i8)') '********CG Solve at timestep ',itime
    write(33,'(a,i6,2e14.6)') &
    'Itn, 2Norm, Rnorm: ',itn,sqrt(rdotr),sqrt(rdotr/rdotr0)
  endif

! CG Iteration
  do
    if(rdotr<=rtol2*rdotr0) then
      if(myrank==0) write(33,*)'JCG converged!'
      exit
    endif
    if(itn>=mxitn) then
      if(myrank==0) write(33,*)'JCG did not converge!'
      exit
    endif

    do ip=1,np 
      if(snz(0,ip)==0) call parallel_abort('JCG: zero diagonal')
      z(ip)=r(ip)/snz(0,ip) !jacobi
    enddo
#ifdef INCLUDE_TIMING
    cwtmp=mpi_wtime()
#endif
    call exchange_p2d(z) !jacobi
#ifdef INCLUDE_TIMING
    wtimer(7,2)=wtimer(7,2)+mpi_wtime()-cwtmp
#endif

    rdotzl=0d0 
    do ip=1,np
      if(associated(ipgl(iplg(ip))%next)) then !interface node
        if(ipgl(iplg(ip))%next%rank<myrank) cycle !already in the sum so skip
      endif
      rdotzl=rdotzl+r(ip)*z(ip)
    enddo
#ifdef INCLUDE_TIMING
    cwtmp=mpi_wtime()
#endif
    call mpi_allreduce(rdotzl,rdotz,1,rtype,MPI_SUM,comm,ierr)
#ifdef INCLUDE_TIMING
    wtimer(7,2)=wtimer(7,2)+mpi_wtime()-cwtmp
#endif
    itn=itn+1
    if(itn==1) then 
      p=z
    else
      if(old_rdotz==0) call parallel_abort('JCG: old_rdotz=0')
      beta=rdotz/old_rdotz 
      p=z+beta*p  !aug. update for p
    endif

    alphal=0d0
    do ip=1,np
      sp(ip)=snz(0,ip)*p(ip)
      do j=1,nnz(ip) 
        sp(ip)=sp(ip)+snz(j,ip)*p(inz(j,ip)) 
      enddo
      if(associated(ipgl(iplg(ip))%next)) then !interface node
        if(ipgl(iplg(ip))%next%rank<myrank) cycle !already in the sum so skip
      endif
      alphal=alphal+p(ip)*sp(ip)
    enddo !ip

#ifdef INCLUDE_TIMING
    cwtmp=mpi_wtime()
#endif
    call exchange_p2d(sp) !update sp ghost values
    call mpi_allreduce(alphal,alpha,1,rtype,MPI_SUM,comm,ierr)
#ifdef INCLUDE_TIMING
    wtimer(7,2)=wtimer(7,2)+mpi_wtime()-cwtmp
#endif

    if(alpha==0) call parallel_abort('JCG: division by zero')
    alpha=rdotz/alpha

    x=x+alpha*p !augmented update
    r=r-alpha*sp !augmented update
    old_rdotz=rdotz 
    rdotrl=0d0
    do ip=1,np 
      if(associated(ipgl(iplg(ip))%next)) then !interface node
        if(ipgl(iplg(ip))%next%rank<myrank) cycle !already in the sum so skip
      endif
      rdotrl=rdotrl+r(ip)*r(ip)
    enddo
#ifdef INCLUDE_TIMING
    cwtmp=mpi_wtime()
#endif
    call mpi_allreduce(rdotrl,rdotr,1,rtype,MPI_SUM,comm,ierr)
#ifdef INCLUDE_TIMING
    wtimer(7,2)=wtimer(7,2)+mpi_wtime()-cwtmp
#endif
    if(mod(itn,moitn)==0.and.myrank==0) write(33,'(a,i6,2e14.6)') &
    'Itn, 2Norm, Rnorm: ',itn,sqrt(rdotr),sqrt(rdotr/rdotr0)
  enddo !CG Iteration

end subroutine solve_jcg

!===============================================================================
!===============================================================================

subroutine tridag(nmax,nvec,n,nc,a,b,c,r,u,gam)
!-------------------------------------------------------------------------------
! This program solves a tridiagonal system. It was adapted from "Numerical 
! Recipes in FORTRAN (pp.43 ).
!
! a,b,c,r: input vectors and are not modified by this program.
! b is the main diagonal, a below and c above. a(1) and c(n) are not used.
! r is the r.h.s.
! (nmax,nvec) is the dimension of r() _and_ u() in the driving routine.
! n: actual rank of the system.
! nc: input; actual # of columns of rhs (<= nvec).
! u: output with nc columns (depending on input nc).
! gam: a working array.
!-------------------------------------------------------------------------------
  use elfe_glbl, only : rkind,errmsg
  use elfe_msgp, only : parallel_abort
  implicit real(rkind)(a-h,o-z), integer(i-n)

  integer, intent(in) :: nmax,nvec,n,nc
  real(rkind), dimension(nmax), intent(in) :: a,b,c
  real(rkind), dimension(nmax,nvec), intent(in) :: r
  real(rkind), dimension(nmax), intent(out) :: gam
  real(rkind), dimension(nmax,nvec), intent(out) :: u

  if(n<1) call parallel_abort('TRIDAG: n must be >= 1')
  if(nc>nvec) call parallel_abort('TRIDAG: Increase # of columns')
  if(b(1)==0d0) call parallel_abort('TRIDAG:  b(1)=0')

  bet=b(1)
  do i=1,nc
    u(1,i)=r(1,i)/bet
  enddo

  do j=2,n
    gam(j)=c(j-1)/bet
    bet=b(j)-a(j)*gam(j)
    if(bet.eq.0d0) call parallel_abort('TRIDAG: failed')
    do i=1,nc
      u(j,i)=(r(j,i)-a(j)*u(j-1,i))/bet
    enddo !i
  enddo !j

! Backsubstitution
  do j=n-1,1,-1
    do i=1,nc
      u(j,i)=u(j,i)-gam(j+1)*u(j+1,i)
    enddo
  enddo
  
end subroutine tridag

